@@ -27,11 +27,12 @@ from dataclasses import astuple, dataclass
 from functools import wraps
 from typing import Dict, List, Optional, Union
 
+from cryptography import x509
+from cryptography.hazmat.primitives.serialization import Encoding
 from lxml.etree import SubElement, _Element
-from OpenSSL.crypto import FILETYPE_ASN1, FILETYPE_PEM, X509, dump_certificate, load_certificate
 
 from .. import SignatureConfiguration, VerifyResult, XMLSignatureProcessor, XMLSigner, XMLVerifier
-from ..algorithms import DigestAlgorithm
+from ..algorithms import DigestAlgorithm, digest_algorithm_implementations
 from ..exceptions import InvalidDigest, InvalidInput
 from ..util import SigningSettings, add_pem_header, ds_tag, namespaces, xades_tag
 
@@ -97,6 +98,8 @@ class XAdESSigner(XAdESProcessor, XMLSigner):
         Parameters to pass to the :class:`signxml.XMLSigner` constructor.
     """
 
+    use_deprecated_legacy_signing_certificate: bool = False
+
     def __init__(
         self,
         signature_policy: Optional[XAdESSignaturePolicy] = None,
@@ -127,7 +130,7 @@ class XAdESSigner(XAdESProcessor, XMLSigner):
         self.namespaces.update(xades=namespaces.xades)
 
     @wraps(XMLSigner.sign)
-    def sign(self, data, always_add_key_value: bool = True, **kwargs) -> _Element:  # type: ignore
+    def sign(self, data, always_add_key_value: bool = True, **kwargs) -> _Element:  # type: ignore[override]
         return super().sign(data=data, always_add_key_value=always_add_key_value, **kwargs)
 
     def _get_token(self, length=4):
@@ -166,7 +169,9 @@ class XAdESSigner(XAdESProcessor, XMLSigner):
         )
         for dop_annotator in self.signed_data_object_properties_annotators:
             dop_annotator(signed_data_object_properties, sig_root=sig_root, signing_settings=signing_settings)
-        self._add_reference_to_signed_info(sig_root, signed_properties, Type="http://uri.etsi.org/01903#SignedProperties")
+        self._add_reference_to_signed_info(
+            sig_root, signed_properties, Type="http://uri.etsi.org/01903#SignedProperties"
+        )
         self._add_reference_to_signed_info(sig_root, key_info)
 
     def _add_reference_to_signed_info(self, sig_root, node_to_reference, **attrs):
@@ -184,21 +189,48 @@ class XAdESSigner(XAdESProcessor, XMLSigner):
     def add_signing_time(self, signed_signature_properties, sig_root, signing_settings: SigningSettings):
         signing_time = SubElement(signed_signature_properties, xades_tag("SigningTime"), nsmap=self.namespaces)
         # TODO: make configurable
-        utc_iso_ts = datetime.datetime.utcnow().isoformat(timespec="seconds")
-        signing_time.text = f"{utc_iso_ts}+00:00"
+        signing_time.text = datetime.datetime.now(datetime.timezone.utc).isoformat(timespec="seconds")
 
     def add_signing_certificate(self, signed_signature_properties, sig_root, signing_settings: SigningSettings):
-        # TODO: check if we need to support SigningCertificate
+        # Emit both legacy SigningCertificate (SHA1 + IssuerSerial) and SigningCertificateV2
         signing_cert_v2 = SubElement(
             signed_signature_properties, xades_tag("SigningCertificateV2"), nsmap=self.namespaces
         )
-        for cert in signing_settings.cert_chain:  # type: ignore
-            if isinstance(cert, X509):
+        assert signing_settings.cert_chain is not None
+        for cert in signing_settings.cert_chain:
+            if isinstance(cert, x509.Certificate):
                 loaded_cert = cert
             else:
-                loaded_cert = load_certificate(FILETYPE_PEM, add_pem_header(cert))
-            der_encoded_cert = dump_certificate(FILETYPE_ASN1, loaded_cert)
+                loaded_cert = x509.load_pem_x509_certificate(add_pem_header(cert))
+            der_encoded_cert = loaded_cert.public_bytes(Encoding.DER)
             cert_digest_bytes = self._get_digest(der_encoded_cert, algorithm=self.digest_alg)
+
+            # Legacy SigningCertificate
+            if self.use_deprecated_legacy_signing_certificate:
+                cert_digest_sha1_bytes = self._get_digest(der_encoded_cert, algorithm=DigestAlgorithm.SHA1)
+                signing_cert = SubElement(
+                    signed_signature_properties, xades_tag("SigningCertificate"), nsmap=self.namespaces
+                )
+
+                cert_node_legacy = SubElement(signing_cert, xades_tag("Cert"), nsmap=self.namespaces)
+                cert_digest = SubElement(cert_node_legacy, xades_tag("CertDigest"), nsmap=self.namespaces)
+                SubElement(
+                    cert_digest, ds_tag("DigestMethod"), nsmap=self.namespaces, Algorithm=DigestAlgorithm.SHA1.value
+                )
+                digest_value_node = SubElement(cert_digest, ds_tag("DigestValue"), nsmap=self.namespaces)
+                digest_value_node.text = b64encode(cert_digest_sha1_bytes).decode()
+                issuer_serial = SubElement(cert_node_legacy, xades_tag("IssuerSerial"), nsmap=self.namespaces)
+                issuer_name = SubElement(issuer_serial, ds_tag("X509IssuerName"), nsmap=self.namespaces)
+                issuer_name.text = "C={C},O={O},OU={OU},CN={CN}".format(
+                    C=loaded_cert.issuer.get_attributes_for_oid(x509.oid.NameOID.COUNTRY_NAME)[0].value,
+                    O=loaded_cert.issuer.get_attributes_for_oid(x509.oid.NameOID.ORGANIZATION_NAME)[0].value,
+                    OU=loaded_cert.issuer.get_attributes_for_oid(x509.oid.NameOID.ORGANIZATIONAL_UNIT_NAME)[0].value,
+                    CN=loaded_cert.issuer.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0].value,
+                )
+                serial_number = SubElement(issuer_serial, ds_tag("X509SerialNumber"), nsmap=self.namespaces)
+                serial_number.text = str(loaded_cert.serial_number)
+
+            # SigningCertificateV2 (current default)
             cert_node = SubElement(signing_cert_v2, xades_tag("Cert"), nsmap=self.namespaces)
             cert_digest = SubElement(cert_node, xades_tag("CertDigest"), nsmap=self.namespaces)
             SubElement(cert_digest, ds_tag("DigestMethod"), nsmap=self.namespaces, Algorithm=self.digest_alg.value)
@@ -276,15 +308,14 @@ class XAdESVerifier(XAdESProcessor, XMLVerifier):
             digest_alg = DigestAlgorithm(self._find(cert_digest, "DigestMethod").get("Algorithm"))
             digest_value = self._find(cert_digest, "DigestValue")
             # check spec for specific method of retrieving cert
-            der_encoded_cert = dump_certificate(FILETYPE_ASN1, expect_cert)
-
-            if b64decode(digest_value.text) != self._get_digest(der_encoded_cert, algorithm=digest_alg):
+            digest_alg_impl = digest_algorithm_implementations[digest_alg]()
+            if b64decode(digest_value.text) != expect_cert.fingerprint(digest_alg_impl):
                 raise InvalidDigest("Digest mismatch for certificate digest")
 
     def _verify_cert_digests(self, verify_result: VerifyResult):
         x509_data = verify_result.signature_xml.find("ds:KeyInfo/ds:X509Data", namespaces=namespaces)
-        cert_from_key_info = load_certificate(
-            FILETYPE_PEM, add_pem_header(self._find(x509_data, "X509Certificate").text)
+        cert_from_key_info = x509.load_pem_x509_certificate(
+            add_pem_header(self._find(x509_data, "X509Certificate").text)
         )
         signed_signature_props = self._find(verify_result.signed_xml, "xades:SignedSignatureProperties")
         signing_cert = self._find(signed_signature_props, "xades:SigningCertificate", require=False)
@@ -331,7 +362,7 @@ class XAdESVerifier(XAdESProcessor, XMLVerifier):
             )
         return self._find(verify_result.signed_xml, "xades:SignedSignatureProperties")
 
-    def verify(  # type: ignore
+    def verify(  # type: ignore[override]
         self,
         data,
         *,
@@ -365,7 +396,7 @@ class XAdESVerifier(XAdESProcessor, XMLVerifier):
             if verify_result.signed_xml is None:
                 continue
             if verify_result.signed_xml.tag == xades_tag("SignedProperties"):
-                verify_results[i] = XAdESVerifyResult(  # type: ignore
+                verify_results[i] = XAdESVerifyResult(  # type: ignore[misc]
                     *astuple(verify_result), signed_properties=self._verify_signed_properties(verify_result)
                 )
                 break
@@ -373,4 +404,4 @@ class XAdESVerifier(XAdESProcessor, XMLVerifier):
             raise InvalidInput("Expected to find a xades:SignedProperties element")
 
         # TODO: assert all mandatory signed properties are set
-        return verify_results  # type: ignore
+        return verify_results  # type: ignore[return-value]
